package app

import (
	"log"
	"os"
	"os/signal"
	"strconv"
	"syscall"

	redis "github.com/go-redis/redis/v8"
	"github.com/jackc/pgx/v5/pgxpool"
	amqp "github.com/rabbitmq/amqp091-go"

	"git.pbiernat.io/egommerce/go-api-pkg/consul"
	"git.pbiernat.io/egommerce/go-api-pkg/fluentd"

	cnf "git.pbiernat.io/egommerce/catalog-service/internal/config"
	srv "git.pbiernat.io/egommerce/catalog-service/internal/server"
	db "git.pbiernat.io/egommerce/catalog-service/pkg/database"
)

type (
	Doer interface {
		Start(while chan struct{}) error
		OnShutdown(*App)
	}
	Application interface {
		Start(while chan struct{})
		RegisterHelper(OptionFn) error
		Shutdown()
	}
	OptionFn func(*App) error

	App struct {
		cnf *srv.ServerConfig

		doer Doer

		// helpers
		Cache    *redis.Client
		Database *pgxpool.Pool
		Eventbus *amqp.Channel
		Logger   *fluentd.Logger
		Registry *consul.Service
	}
)

func NewApp(c *cnf.ServerConfig, d Doer) *App {
	app := &App{
		cnf:  c,
		doer: d,
	}

	// while := make(chan struct{})
	// // err := srv.Base.Start(while)
	// err := app.Doer.Start(while)
	// <-while

	// if err != nil {
	// 	log.Fatalf("Failed to start application. Reason: %v\n", err)
	// 	os.Exit(1)
	// }

	// os.Exit(0)

	return app
}

func (a *App) Start(while chan struct{}) error {
	go func() {
		sigint := make(chan os.Signal, 1)
		signal.Notify(sigint, os.Interrupt, syscall.SIGINT, syscall.SIGTERM)
		<-sigint

		a.doer.OnShutdown(a)
		a.Shutdown()

		close(while)
	}()

	run := a.createRunFile("./app.run") // FIXME path...
	defer a.removeRunFile(run)

	a.doer.Start(while)
}

func (a *App) RegisterHelper(name string, opt OptionFn) error {
	if err := opt(a); err != nil {
		log.Fatalf("Failed to attach extension to the server. Err: %v\n", err)
	}
}

func (a *App) Shutdown() {
	// ...
}

func (a *App) createRunFile(path string) *os.File {
	run, err := os.Create(path)
	if err != nil {
		log.Fatalf("Failed to create run file. Reason: %v\n", err)
		os.Exit(1)
	}
	run.WriteString(strconv.Itoa(os.Getpid()))

	return run
}

func (a *App) removeRunFile(f *os.File) error {
	return f.Close()
}

func WithCache() OptionFn {
	return func(a *App) error {
		a.Cache = redis.NewClient(&redis.Options{
			Addr:     a.cnf.CacheAddr,
			Password: a.cnf.CachePassword,
			DB:       0,
		})

		return nil
	}
}

func WithDatabase() OptionFn {
	return func(a *App) error {
		dbConn, err := db.Connect(a.cnf.DbURL)
		if err != nil {
			log.Fatalf("Failed to connect to the Database: %s. Err: %v\n", a.cnf.DbURL, err)
			os.Exit(1) // TODO: retry in background... for now treat as integraded service...
		}

		a.Database = dbConn

		return nil
	}
}

func WithEventbus() OptionFn {
	return func(a *App) error {
		conn, err := amqp.Dial(a.cnf.EventBusURL)
		if err != nil {
			log.Fatalf("Failed to connect to the Eventbus: %s. Err: %v\n", a.cnf.EventBusURL, err)
		}

		chn, err := conn.Channel()
		if err != nil {
			log.Fatalf("Failed to open new Eventbus channel. Err: %v\n", err)
		}

		a.Eventbus = chn

		return nil
	}
}

func WithLogger() OptionFn {
	return func(a *App) error {
		logHost, logPort, err := fluentd.ParseAddr(a.cnf.LoggerAddr)
		if err != nil {
			log.Fatalf("Failed to parse Fluentd address: %s. Err: %v", a.cnf.LoggerAddr, err)
		}

		logger, err := fluentd.NewLogger(a.cnf.Base.GetAppFullName(), logHost, logPort)
		if err != nil {
			log.Fatalf("Failed to connect to the Fluentd on %s:%d. Err: %v", logHost, logPort, err)
		}

		a.Logger = logger

		return nil
	}
}

func WithRegistry() OptionFn {
	return func(a *App) error {
		port, _ := strconv.Atoi(a.cnf.Base.NetAddr[1:]) // FIXME: can be IP:PORT or :PORT
		// log.Printf("Consul retrieved port: %v", port)
		registry, err := consul.NewService(a.cnf.RegistryAddr, a.cnf.Base.AppID, a.cnf.Base.AppName, a.cnf.Base.RegistryDomainOverIP, a.cnf.Base.GetIP(), a.cnf.Base.AppDomain, a.cnf.Base.PathPrefix, port)
		if err != nil {
			log.Fatalf("Failed to connect to the Consul on: %s. Err: %v", c.RegistryAddr, err)
		}

		err = registry.Register()
		if err != nil {
			log.Fatalf("Failed to register in the Consul service. Err: %v", err)
		}

		registry.RegisterHealthChecks()
		// a.registerKVUpdater() // FIXME run as goroutine

		a.Registry = registry

		// svc, _ := registry.Connect()
		// tlsCnf := svc.ServerTLSConfig()
		// s.Base.App.Server().TLSConfig = tlsCnf
		// fmt.Println("Podmiana configa TLS")
		// defer svc.Close()

		// go func() { // Consul KV updater
		// 	ticker := time.NewTicker(time.Second * 15)
		// 	for range ticker.C {
		// 		fetchKVConfig(s) // FIXME: duplicated in worker
		// 	}
		// }()

		// go func() { // Server metadata cache updater
		// 	ticker := time.NewTicker(time.Second * 5)
		// 	for range ticker.C {
		// 		s.cacheMetadata()
		// 	}
		// }()

		return nil
	}
}
